// fpga_dll.cpp : FPGA-based Fast Multisource Pulse Registration System 
// Simple Labview Driver
// by Sergey Polyakov

// The computer communicates with the FPGA via a USB 2 port and using a standard ezusb driver.
// The FPGA is a slave USB device. The protocol consists of 
// 1) a request sent by a computer (sends one byte via FIFO2, which is used for a primitive control of FPGA, see documentation)
// 2) a response of FPGA (replies with 2 bytes via FIFO5, reports the length of collected information in bytes)
// 3) data of a total length reported in step 2 is then availible for download via FIFO4

// This dll computes simple statistics of single pulses and pulse coincidences and provides for an option to write all timestamps and events.
// It provides a simple iterface for labview.

// Interface functions:
// HANDLE FPGA_Open() initializes the FPGA communication protocol. 
//		returns an integer "handle" to the device. -1 if not successful.
//
// void FPGA_Close(HANDLE XyloDeviceHandle) Closes communication session.
//
// int FPGA_Interface (int runs, char fpga_command, int saveclicks, char * filename, HANDLE XyloDeviceHandle, int * stats)
//		establishes the communication with the FPGA and returns statistics.
//              int runs: query the FPGA "runs" number of times, and accumulate statistics
//              char fpga_command: one byte to send to FPGA. Possible commands: 0b00000000 no command
//																				0b00000001 clear timestamping counter (same as start)
//																				0b00000010 stop aquisition
//																				0b00000100 start aquisition
//				int saveclicks: if =1, save timestamped electronic events in a file; =0 do not save
//				char * filename: name of file to save events into
//				HANDLE XyloDeviceHandle: handle to FPGA device
//				int *stats: pointer to the output statistics array
//		returns 1 if successful, 0 upon an error
#include "stdafx.h"

//Name of a function with real time access to received data.
//User could write his/her own function and give its name here. Default function is "correlate"
//The user function must be declared as
//void <function_name> ( unsigned char * data_in, int data_length, int *stats)
//where data_in is an array of recorded events, 4 bytes are used per one event
//data_length is the length of the array (divide by 4 to get number of events)
//stats is an array of integers, with any user output. The pointer to zeroth element will be passed to labview. Labview code needs to know the length of this array.


#define REALTIME_FUNCTION correlate


///////////////////////////////////////////////////
// Default real-time
// Simple correlation analysis. Computes statistics.
///////////////////////////////////////////////////
/*
stats structure: number of occurences of these events

 stats[0] state 0
 stats[1] state 1
 stats[2] state 2
 stats[3] state 3
 stats[4] states 0&1
 stats[5] states 0&2
 stats[6] states 0&3
 stats[7] state 1&2
 stats[8] state 1&3
 stats[9] state 2&3 
 stats[10] states 0&1&2
 stats[11] states 0&1&3
 stats[12] states 0&2&3
 stats[13] state 1&2&3              
 stats[14] states 0&1&2&3

*/
void correlate ( unsigned char * data, int length, int *stats)
{
    int i;
    int dlength = length%4;
    
    for (i=0; i<length; i+=4)
    {
        int state[4];
        state[0] = (data[i+3]&0x08)?true:false;
        state[1] = (data[i+3]&0x10)?true:false;
        state[2] = (data[i+3]&0x20)?true:false;
        state[3] = (data[i+3]&0x40)?true:false;

        stats[2] += (int)state[2];//state 2
        stats[3] += (int)state[3];//state 3
        
        if (state[0])
        {
                stats[0] += 1; //state 0;
     
                stats[4] += (state[1])?1:0; //states 0&1
                stats[5] += (state[2])?1:0; //states 0&2
                stats[6] += (state[3])?1:0; //states 0&3
                
                stats[10] += (state[1]&state[2])?1:0; //states 0,1&2
                stats[11] += (state[1]&state[3])?1:0; //states 0,1&3
                stats[12] += (state[2]&state[3])?1:0; //states 0,2&3

                stats[14] += (state[1]&state[2]&state[3])?1:0; //states 0,1,2&3
        }
         
        if (state[1])
        {        
                 stats[1] += 1;//state 1
                 
                 stats[7] += (state[2])?1:0; //state 1&2
                 stats[8] += (state[3])?1:0; //state 1&3
        
                 stats[13] += (state[2]&state[3])?1:0;  //state 1&2&3              
        }
        stats[9] += (state[2]&state[3])?1:0; //state 2&3
    }
    return;             
}


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
// Defines the entry point for the DLL application.					 
{
    return TRUE;
}


///////////////////////////////////////////////////
// Opens and closes the FPGA/USB driver
///////////////////////////////////////////////////
_declspec(dllexport) HANDLE FPGA_Open()
{
    HANDLE XyloDeviceHandle;
	XyloDeviceHandle = CreateFile("\\\\.\\EzUSB-0", GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
	if (XyloDeviceHandle==INVALID_HANDLE_VALUE) return (HANDLE)-1;
	return XyloDeviceHandle;
}

_declspec(dllexport) void FPGA_Close(HANDLE XyloDeviceHandle)
{
	CloseHandle(XyloDeviceHandle);
}
///////////////////////////////////////////////////
// FPGA/USB interface (raw functions)
///////////////////////////////////////////////////
int USB_BulkWrite(ULONG pipe, void* buffer, ULONG buffersize, HANDLE XyloDeviceHandle)
{
	DWORD nBytes;
	int result; 
    result = 
           DeviceIoControl(XyloDeviceHandle, 0x222051, &pipe, sizeof(pipe), buffer, buffersize, &nBytes, NULL);
    if (nBytes== buffersize && result==1)
    return 1; //ok
    else
    return -1; //error
}

DWORD USB_BulkRead(ULONG pipe, void* buffer, ULONG buffersize, HANDLE XyloDeviceHandle)
{
	DWORD nBytes;
	int result;
	result = DeviceIoControl(XyloDeviceHandle, 0x22204E, &pipe, sizeof(pipe), buffer, buffersize, &nBytes, NULL);
//	bool DIC (...) If the operation completes successfully, the return value is nonzero.
//  If the operation fails or is pending, the return value is zero. To get extended error information, call GetLastError.

	if (result !=1 ) return -1; //error
	return nBytes;
}


///////////////////////////////////////////////////
// Communicates with FPGA
///////////////////////////////////////////////////
_declspec(dllexport) int FPGA_Interface (int runs, char fpga_command, int saveclicks, char * filename, HANDLE XyloDeviceHandle, int * stats)
{
    FILE *clicklog;
    unsigned char buffer[65536];
    unsigned char*nbuffer;
    const int BSize = 0x8000;
    
    int i;

	// if communication resulted in less bytes than is necessary to define an event (4), the portion of the last event are saved for later    
    static int moredataflag=0; 
    static int moredata[4];
	// if the number of bytes availible for downloading is greater than we are ready to accept, they will be left until the next readout
	static unsigned int bufferleft=0;
	int shortpacketflag;
    
    
    for (i=0; i<15; i++ )
    {
        stats[i] = 0;
    }
    if (saveclicks)
    {
     clicklog = fopen(filename, "at");
    }
    
    int ct = time(NULL); //timeout check: if data aquisition is too slow this finishes the routine without running the protocol "runs" times
    for (i=0; i<runs; i++)
    {
		//step 1: use bytes downloaded previously, that did not define the event.
        if (moredataflag)
        {
                         int kk;
                         for (kk=0; kk<moredataflag; kk++)
                         {
                             buffer[kk] = moredata[kk];
                         }
                         nbuffer = buffer + moredataflag;

        }
        else nbuffer = buffer;
        //step 2: write a command into FPGA and request data
        int nb_bytes_received;
        if ( USB_BulkWrite(2, &fpga_command, 1, XyloDeviceHandle)== -1 ) 
        {
             return 0; //error
        }
		//step 3: recieve 2 bytes defining the length of transmission
		if ( (nb_bytes_received = USB_BulkRead(5, nbuffer, 512, XyloDeviceHandle) )== -1)
		{
			return 0; //error
		}
		//calculate the length of data availible for download
		unsigned int size = (int)nbuffer[1]+256*(int)nbuffer[0];
		//because a full USB frame is 512 bytes, download an integer number of frames, and leave everthing else for a next session
		size += bufferleft;
		bufferleft = size%512;
		size -=bufferleft;	
		//however, if there was too little data availible (i.e. less than <512 bytes), the FPGA will force sending a short packet to keep us informed
		shortpacketflag = 0;
		if (!size) 
		{
			shortpacketflag = 1;
			size = 1024;
		}
		//step 4: recieve a bulk of data
        if ( (nb_bytes_received = USB_BulkRead(4, nbuffer, size, XyloDeviceHandle) )== -1) 
	    {
		     return 0; //error
		}  
		if ( shortpacketflag )
		{
			shortpacketflag = 0;
			bufferleft -= nb_bytes_received;
		}		
		//step 5: check if the information on the last event has to be stored for later
        if (moredataflag)
	    {
			nb_bytes_received += moredataflag;
	        moredataflag=0;
	    }
		if ( (moredataflag=nb_bytes_received%4)!=0)
	    {
                                  int k;
                                  for (k=0; k<moredataflag; k++)
                                  {
                                      moredata[k] = buffer[nb_bytes_received-moredataflag+1+k];
                                  }
                                  nb_bytes_received -= moredataflag;
		}

		//step 6: save a file, if needed
		if (saveclicks)
		{
                      int j;
                      for (j=0; j<nb_bytes_received; j+=4)
                      {
                          fprintf(clicklog,"%X.%X.%X.%X\n",buffer[j+3], buffer[j+2], buffer[j+1], buffer[j]);
                      }
		              fflush(clicklog);
		}
		//step 7: compute the simple statistics
		REALTIME_FUNCTION (buffer,nb_bytes_received,stats);
		if ( ct!=(int)time(NULL) ) break;
    }
    return 1; //no error
}
